!===============================================================================!
! yage.h - Yet Another Glulx Extension
!===============================================================================!
! by Paolo Lucchesi
! with a lot of help from Alessandro Schillaci
! also thanks to Paolo Maroncelli & Vincenzo Scarpa
!
! WeB   : http://www.paololucchesi.it/at
! @mail : pl@paololucchesi.it
!
! Version 0.9 2010-01-31
!===============================================================================!
! What is this?
! -------------
! yage.h (this file) is an inform-glulx extension, a set of functions
! to manage images and music with little effort.
! The graphic functions are similar to the ones used in "Beyond" and
! "La Pietra della Luna", with images shown on the top, on the left or
! on the right of the main window.
! The sound functions can be used to play music, and to fade from
! a song to the next one (or simply fade out).
!===============================================================================!
! yage.h and sgw.h
! ----------------
! The first idea was to integrate yage.h functions in sgw.h, but the
! windows management was rather different, so that it wasn't possible
! to do it without breaking sgw.h backward compatibility.
! It shouldn't be too difficult, for an expert programmer, to mix parts
! of yage.h and sgw.h to make a custom glulx extension.
! There are a few differences between yage.h and sgw.h.
! Graphics: sgw.h has a fixed size graphic window on the top, while yage.h
! opens dinamically the graphic window according to the image size (and it
! can put it on the left or on the right).
! Styles: sgw.h uses a few style wrappers, like s_bold and s_underline,
! while yage.h use a single style function, SetStyle, which is similar
! to the canonical inform style statement.
! Sounds: both libraries have similar functions, but sgw.h can handle
! two additional sound channels, while yage.h can't.
!===============================================================================!
! How to use yage.h
! -----------------
! Put in your main file the 
!   include "yage"
! directive, after including "parser" and "infglk".
!
! Define images and sound (in this order) in your resource file.
!
! Use the functions to show images, switch styles and play sounds.
!
! You can customize colors and other details defining a few constant
! before including "yage":
!
! SCBACK        Main window and graphic window background color
! SCTEXT        Main window text color
! SCSOFT        Main window 'soft' color
! SCEMPH        Main window emphatized/bold color
! SCHEAD        Main window header color
! SCINPU        Main window inputline color
! SLBACK        Statusline background color
! SLTEXT        Statusline text color
! SLEMPH        Statusline emphatized color
! IMGBORDER     Border around images, in pixels (default 0)
! IMGTOPGAP     Maximum gap above image in left and right window
!               Set it to -1 to center image vertically (default 32)
! MAXIMAGES     Max number of images (for conditional images, default 64)
! MAXVOLUME     Music volume (default 16384)
! DECVOLUME     Fade out time, in 1/50 sec (default 64)
!
! You can use the $rrggbb format for specify hexadecimal rgb color codes.
!
! You can also define a NOSOUND constant to avoid sound entirely.
!
! You may want to define your constants between #IfDef TARGET_GLULX; and
! #EndIf; to avoid warnings.
!
! All functions are wrapped so that they can be used, with no effect, in
! z-code programs too.
!
!===============================================================================!
! Images
! ------
! Graphic window size depends on image size (and IMGBORDER value).
! Images in left and right windows are placed near the top (depending
! on IMGTOGAP value) or centered (if IMGTOGAP is negative).
! Images in above window is always centered.
! Furthermore image is not shown if it cannot fit in window, or if it
! take a too large portion of the main window.
! Images are rearranged (and eventually removed or redrawn) if the player
! resizes the game window.
!
! LeftPicture(pic)      Show image on the left of the main window
! RightPicture(pic)     Show image on the right of the main window
! AbovePicture(pic)     Show image above the main window
! NoPicture(pic)        Remove image and maximize the main window
! NoPictureNewLoc(pic)  Remove image only if player location has changed
!===============================================================================!
! Conditional Images
! ------------------
! Conditional images are picture shown only the first time (unless yage.h
! is told otherwise). Be sure that MAXIMAGES is greater than the number of
! images you use, and that images are declared before sounds in resource
! file.
!
! LeftPictureChech(pic)   As LeftPicture, but only if pic wasn't shown before
! RightPictureChech(pic)  As RightPicture, but only if pic wasn't shown before
! AbovePictureChech(pic)  As AbovePicture, but only if pic wasn't shown before
! CheckImage(pic)         Return true only if pic wasn't shown before
! SetImage(pic)           Set pic as "shown before" - return true if it wasn't
! ResetImage(pic)         Set pic as "not shown before"
!===============================================================================!
! Styles
! ------
! There's a single style function, SetStyle, which take a single argument:
! SetStyle(SNORMAL)     Normal typeface (as inform 'style roman' statement)
! SetStyle(SROMAN)      As above
! SetStyle(SBOLD)       Bold typeface (as inform 'style bold' statement)
! SetStyle(SITALIC)     Italic typeface (as inform 'style underline' statement
!                       in z-code, but a true italic in glulx)
! SetStyle(SSOFT)       Soft typeface (no effect in z-code)
! SetStyle(SREVERSE)    Reversed typeface (as inform 'style reverse' statement)
!
! You can use SetStyle as a printing rule, but remember that you have to
! reset manually the normal style:
!
! print "You can see a ", (SetStyle)SITALIC, "strange", (SetStyle)SNORMAL,
!       " and " (SetStyle)SBOLD, "dark", (SetStyle)SNORMAL, " mist.";
!===============================================================================!
! Sounds
! ------
! There are only four functions to manage music, plus other three to turn
! all sounds on or off. The 'restart' parameter, if true, force the
! cut or fade even if the same song is already playing.
! 
! PlayMusic(snd,restart)    Play snd directly
! FadeToMusic(snd,restart)  Fade out current song, and then play snd
! StopMusic()               Stop current song immediately
! FadeOutMusic()            Fade out current song
! 
! MusicOff()                Fade out current song and turn all sounds off
! MusicOn()                 Turn sounds on
! MusicSwitch()             Toggle all sounds on and off. Return true if 
!                           sounds are now on.
!===============================================================================!

System_file;

Constant SNORMAL 0;
Constant SROMAN 0;
Constant SBOLD 1;
Constant SITALIC 2;
Constant SSOFT 3;
Constant SREVERSE 4;

#IfDef TARGET_GLULX;

Constant GG_GXWIN_ROCK 210;

Global gg_gxwin = 0;

Object "(yage)" LibraryExtensions
  with ext_initialise [;yagemain(); ];

Default SCBACK $ffffff;
Default SCTEXT $111111;
Default SCSOFT $888888;
Default SCEMPH $555555;
Default SCHEAD $444444;
Default SCINPU $555555;
Default SLBACK $dddddd;
Default SLTEXT $222222;
Default SLEMPH $444444;
Default IMGBORDER 0;
Default IMGTOPGAP 32;

[ InitGlkWindow winrock;
    switch (winrock) {
      GG_MAINWIN_ROCK:
        glk_stylehint_set(wintype_AllTypes, style_Normal, stylehint_TextColor, SCTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Normal, stylehint_BackColor, SCBACK);
        glk_stylehint_set(wintype_AllTypes, style_Normal, stylehint_Justification, stylehint_just_LeftRight);        
        glk_stylehint_set(wintype_AllTypes, style_Emphasized, stylehint_TextColor, SCEMPH);
        glk_stylehint_set(wintype_AllTypes, style_Emphasized, stylehint_BackColor, SCBACK);
        glk_stylehint_set(wintype_AllTypes, style_Header, stylehint_TextColor, SCHEAD);
        glk_stylehint_set(wintype_AllTypes, style_Header, stylehint_BackColor, SCBACK);
        glk_stylehint_set(wintype_AllTypes, style_Subheader, stylehint_TextColor, SCEMPH);
        glk_stylehint_set(wintype_AllTypes, style_Subheader, stylehint_BackColor, SCBACK);
        glk_stylehint_set(wintype_AllTypes, style_Alert, stylehint_TextColor, SCBACK);
        glk_stylehint_set(wintype_AllTypes, style_Alert, stylehint_BackColor, SCTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Note, stylehint_TextColor, SCINPU);
        glk_stylehint_set(wintype_AllTypes, style_Note, stylehint_BackColor, SCBACK);       
        glk_stylehint_set(wintype_AllTypes, style_Blockquote, stylehint_TextColor, SCINPU);
        glk_stylehint_set(wintype_AllTypes, style_Blockquote, stylehint_BackColor, SCBACK); 
        glk_stylehint_set(wintype_AllTypes, style_Input, stylehint_TextColor, SCINPU);
        glk_stylehint_set(wintype_AllTypes, style_Input, stylehint_BackColor, SCBACK);
        glk_stylehint_set(wintype_AllTypes, style_Preformatted, stylehint_TextColor, SCINPU);
        glk_stylehint_set(wintype_AllTypes, style_Preformatted, stylehint_BackColor, SCBACK);  
        glk_stylehint_set(wintype_AllTypes, style_User1, stylehint_TextColor, SCSOFT);      
        glk_stylehint_set(wintype_AllTypes, style_User1, stylehint_BackColor, SCBACK);  
        glk_stylehint_set(wintype_AllTypes, style_User2, stylehint_TextColor, SCTEXT);
        glk_stylehint_set(wintype_AllTypes, style_User2, stylehint_BackColor, SCBACK);
        glk_stylehint_set(wintype_AllTypes, style_User2, stylehint_Justification, stylehint_just_LeftRight);
        glk_stylehint_set(wintype_AllTypes, style_User2, stylehint_Oblique, true);  
      GG_STATUSWIN_ROCK:
        glk_stylehint_set(wintype_AllTypes, style_Normal, stylehint_TextColor, SLTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Normal, stylehint_BackColor, SLBACK);
        glk_stylehint_set(wintype_AllTypes, style_Normal, stylehint_Justification, stylehint_just_LeftRight);        
        glk_stylehint_set(wintype_AllTypes, style_Emphasized, stylehint_TextColor, SLEMPH);
        glk_stylehint_set(wintype_AllTypes, style_Emphasized, stylehint_BackColor, SLBACK);
        glk_stylehint_set(wintype_AllTypes, style_Header, stylehint_TextColor, SLEMPH);
        glk_stylehint_set(wintype_AllTypes, style_Header, stylehint_BackColor, SLBACK);
        glk_stylehint_set(wintype_AllTypes, style_Subheader, stylehint_TextColor, SLEMPH);
        glk_stylehint_set(wintype_AllTypes, style_Subheader, stylehint_BackColor, SLBACK);
        glk_stylehint_set(wintype_AllTypes, style_Alert, stylehint_TextColor, SLBACK);
        glk_stylehint_set(wintype_AllTypes, style_Alert, stylehint_BackColor, SLTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Note, stylehint_TextColor, SLTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Note, stylehint_BackColor, SLBACK);       
        glk_stylehint_set(wintype_AllTypes, style_Blockquote, stylehint_TextColor, SLTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Blockquote, stylehint_BackColor, SLBACK); 
        glk_stylehint_set(wintype_AllTypes, style_Input, stylehint_TextColor, SLTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Input, stylehint_BackColor, SLBACK);
        glk_stylehint_set(wintype_AllTypes, style_Preformatted, stylehint_TextColor, SLTEXT);
        glk_stylehint_set(wintype_AllTypes, style_Preformatted, stylehint_BackColor, SLBACK);  
        glk_stylehint_set(wintype_AllTypes, style_User1, stylehint_TextColor, SLEMPH);      
        glk_stylehint_set(wintype_AllTypes, style_User1, stylehint_BackColor, SLBACK);         
    }
    rfalse; 
];
   
[ yagemain;
    TestGlkCapabilities();
    MeasureWindow();
#Ifndef NOSOUND;
    TimerInit();
    SoundInit();
#Endif;     
];

! --- Capabilities testing

Global gs_graphic=false;
Global gs_timer=false;
!Global gs_imgfade=false;
#IfnDef NOSOUND;
Global gs_music=false;
Global gs_musicfade=false;
#Endif;

[ TestGlkCapabilities;
  if (glk_gestalt(gestalt_Graphics)) gs_graphic=true;
  if (glk_gestalt(gestalt_Timer)) gs_timer=true;
  !if (gs_graphic && gs_timer && glk_gestalt(gestalt_GraphicsTransparency)) gs_imgfade=true;
#IfnDef NOSOUND;
  if (glk_gestalt(gestalt_Sound)) gs_music=true;
  if (gs_music && gs_timer && glk_gestalt(gestalt_SoundVolume)) gs_musicfade=true;
#Endif;
];

! --- Styles wrapper

[ SetStyle x;
    switch(x) {
      SROMAN: glk_set_style(style_Normal);
      SBOLD: glk_set_style(style_Subheader);
      SITALIC: glk_set_style(style_User2);
      SSOFT: glk_set_style(style_User1);
      SREVERSE: glk_set_style(style_Alert);
    }
];

! --- Graphics

Global gxwinside=0;
Global gxwinmeth=0;
Global gxwindim=0;
Global imgx=0;
Global imgy=0;
Global gxpicture=0;
Global gxlocation=0;
Global picx;
Global picy;
Global winx;
Global winy;
   
[ CloseWindow;
    if (~~gs_graphic) rfalse;
    glk_window_close(gg_gxwin, 0);
    gg_gxwin = 0;
];

[ MeasureWindow;
  if (~~gs_graphic) rfalse;
  CloseWindow();
  gg_gxwin = glk_window_open(gg_mainwin, winmethod_Above+winmethod_Proportional, 100, wintype_Graphics, GG_GXWIN_ROCK);
  glk_window_get_size(gg_gxwin, gg_arguments, gg_arguments+4);
  winx = gg_arguments-->0;
  winy = gg_arguments-->1;   
  CloseWindow();
];

[ ResizeWindow;
  MeasureWindow();
  MakeWindow();
];

[ MakeWindow pwin;
  if (~~gs_graphic) rfalse;
  !print " gxpicture=",gxpicture," gxwindim=",gxwindim, " gxwinside=",gxwinside, " gxwinmeth=",gxwinmeth;
  !print " winx=",winx," winy=",winy," imgx=",imgx," imgy=",imgy,"^";
  if ((~~gxpicture) || (~~gxwindim)) { CloseWindow(); rtrue; }
  if ((gxwinside == winmethod_Above) && ((5*gxwindim) > (2*winy))) { CloseWindow(); rtrue; }
  if ((gxwinside == winmethod_Above) && (imgx > winx)) { CloseWindow(); rtrue; }
  if ((gxwinside == winmethod_Left or winmethod_Right) && ((3*gxwindim) > winx)) { CloseWindow(); rtrue; }
  if ((gxwinside == winmethod_Left or winmethod_Right) && (imgy > winy)) { CloseWindow(); rtrue; }
  if (gg_gxwin) {
      pwin=glk_window_get_parent(gg_gxwin); 
      glk_window_get_arrangement(pwin, gg_arguments, gg_arguments+4, gg_arguments+8); 
      !print "Now method=",gg_arguments-->0," size=",gg_arguments-->4,"^"; keycharprimitive();
      if ((gg_arguments-->0) == (gxwinside+gxwinmeth)) 
            glk_window_set_arrangement(pwin, gxwinside+gxwinmeth, gxwindim, gg_gxwin);
      else
            CloseWindow();
  }
  if (~~gg_gxwin) gg_gxwin = glk_window_open(gg_mainwin, gxwinside+gxwinmeth, gxwindim, wintype_Graphics, GG_GXWIN_ROCK);
  glk_window_set_background_color(gg_gxwin, SCBACK);
  glk_window_clear(gg_gxwin); 
  RedrawGraphicImage();
];

[ RedrawGraphicImage wx wy ix iy px py;
    !SetCenteredPosition();
    if ((~~gg_gxwin) || (~~gxpicture) || (~~gs_graphic)) rfalse;
    glk_window_get_size(gg_gxwin, gg_arguments, gg_arguments+4);
    wx = gg_arguments-->0; wy = gg_arguments-->1;
    glk_image_get_info(gxpicture, gg_arguments, gg_arguments+4);
    ix = gg_arguments-->0; iy = gg_arguments-->1;      
    py = (wy-iy)/2;
    if (py < 0) py = 0;
    if ((gxwinside == winmethod_Left or winmethod_Right) && (IMGTOPGAP >= 0) && (py > IMGTOPGAP+IMGBORDER)) py=IMGTOPGAP+IMGBORDER;
    px = (wx-ix)/2;
    if (px < 0) px = 0;
    glk_image_draw(gg_gxwin, gxpicture, px, py);
];

Default MAXIMAGES 64;
Array imgused->(MAXIMAGES+8);

[ CheckImage pic; if (imgused->pic) rfalse; rtrue; ];  
[ SetImage pic; if (imgused->pic) rfalse; imgused->pic = true; rtrue; ];  
[ ResetImage pic; imgused->pic = false; rtrue; ];

[LeftPictureCheck pic;
    if (imgused->pic) rfalse; imgused->pic = true;
    LeftPicture(pic);
];

[RightPictureCheck pic;
    if (imgused->pic) rfalse; imgused->pic = true;
    RightPicture(pic);
];

[AbovePictureCheck pic;
    if (imgused->pic) rfalse; imgused->pic = true;
    AbovePicture(pic);
];

[LeftPicture pic;
    if (~~gs_graphic) rfalse;
    gxwinside = winmethod_Left;
    gxwinmeth = winmethod_Fixed;
    gxpicture = pic;
    gxlocation = real_location;
    glk_image_get_info(gxpicture, gg_arguments, gg_arguments+4);
    imgx = gg_arguments-->0;
    imgy = gg_arguments-->1;
    gxwindim = imgx+2*IMGBORDER;
    MakeWindow();
];

[RightPicture pic;
    if (~~gs_graphic) rfalse;
    gxwinside = winmethod_Right;
    gxwinmeth = winmethod_Fixed;
    gxpicture = pic;
    gxlocation = real_location;
    glk_image_get_info(gxpicture, gg_arguments, gg_arguments+4);
    imgx = gg_arguments-->0;
    imgy = gg_arguments-->1;
    gxwindim = imgx+2*IMGBORDER;    
    MakeWindow();
];

[AbovePicture pic;
    if (~~gs_graphic) rfalse;
    gxwinside = winmethod_Above;
    gxwinmeth = winmethod_Fixed;    
    gxpicture = pic;
    gxlocation = real_location;
    glk_image_get_info(gxpicture, gg_arguments, gg_arguments+4);
    imgx = gg_arguments-->0;
    imgy = gg_arguments-->1;
    gxwindim = imgy+2*IMGBORDER;
    MakeWindow();
];    

[ NoPictureNewLoc;
    if (gxlocation == real_location or 0) rfalse;
    NoPicture();
];


[ NoPicture;
    if (~~gs_graphic) rfalse;
    if (gxpicture == 0) rfalse;
    gxwinside = 0;
    gxwinmeth = 0;
    gxwindim = 0;
    gxpicture = 0;
    gxlocation = 0;
    MakeWindow();
];
  
!--- Music

#Ifndef NOSOUND;
Constant GG_MUSICCHAN_ROCK 410;
Global gg_musicchan = 0; 

Global timer_event = 0;
Global timer_time = 0;
  
Default MAXVOLUME $4000;
Default DECVOLUME $40;

Global musicflag = true;
Global current_music = 0;
Global next_music = 0;

Global music_volume = MAXVOLUME;
Global fade_music = false;

[ TimerInit;
   if (~~gs_timer) rfalse;
   glk_request_timer_events(50);
];

[ SoundInit;
   if (~~gs_music) rfalse;
   if (gg_musicchan == 0) {
     gg_musicchan = glk_schannel_create(GG_MUSICCHAN_ROCK);
   }
];

[ MusicSwitch;
    if (~~gs_music) rfalse;
    if (musicflag) { MusicOff(); rfalse; }
     else          { MusicOn(); rtrue; }
];
    
[ MusicOff;
    if (~~gs_music) rfalse;
    if (musicflag) {
      musicflag = false;
      FadeOutMusic();
    }  
];

[ MusicOn;
    if (~~gs_music) rfalse;
    musicflag = true;
    !MyRestartMusicChannel();
];

[ MyRestartMusicChannel;
   if (~~gs_music) rfalse;
   if (gg_musicchan) {
     if (current_music == 0) glk_schannel_stop(gg_musicchan);
     else glk_schannel_play_ext(gg_musicchan, current_music, 1, 1);
     glk_schannel_set_volume(gg_musicchan, music_volume);
   }
];

[ StopMusic;
    if (~~gs_music) rfalse;
    current_music = 0;
    MyRestartMusicChannel();
];

[ FadeOutMusic;
    if (~~gs_music) rfalse;
    if (~~gs_musicfade) return(StopMusic());
    fade_music = true;
    next_music = 0;
];

[ FadeToMusic song restart;
    if (~~gs_music) rfalse;
    if (~~gs_musicfade) return(PlayMusic(song,restart));
    !print "Fade to music ", current_music, " -> ", song, ".^";
    if (musicflag) {
      if ((song == current_music) && (~~fade_music) && (~~restart)) rtrue;
      if (current_music) {
          FadeOutMusic();
          next_music = song;
          rtrue;
      }
      PlayMusic(song,restart);
    }
];    
    
[ PlayMusic song restart;
    if (~~gs_music) rfalse;
    !print "Play music ", song, ".^";
    if (musicflag) {
      if ((song == current_music) && (~~restart)) rtrue;
      StopMusic();
      if (song == 0) rtrue;
      current_music = song;
      music_volume = MAXVOLUME;
      MyRestartMusicChannel();
    }
];

[ DecreaseVolume;
    if (~~gs_music) rfalse;
    !print music_volume, "-", fade_music, "*";
    music_volume = music_volume - MAXVOLUME / DECVOLUME;
    if (music_volume < 0) music_volume = 0;
    if (gg_musicchan) glk_schannel_set_volume(gg_musicchan, music_volume);
    if (music_volume == 0) {
        fade_music = false;
        current_music = 0;
        PlayMusic(next_music);
        next_music = 0;
    }
];

#Endif;

[ HandleGlkEvent ev context x;
  x = context;
  switch (ev-->0) {
    evtype_Timer:
    #Ifndef NOSOUND;
      timer_time = timer_time + 50;
      if (timer_event) timer_event();
      if (fade_music) DecreaseVolume();
    #Endif;
    evtype_Arrange, evtype_Redraw:
      ResizeWindow();
  }
  ev=ev;
];

[ IdentifyGlkObject phase type ref rock res id;
    res = res;
    if (phase == 0) { ! Zero out references to our objects.
  #Ifndef NOSOUND;    
        gg_musicchan = 0; 
  #Endif;
        gg_gxwin = 0;
    }
    if (phase == 1) { ! Reset our windows, streams and filerefs.
        switch (type) {
          0: switch (rock) {
               GG_GXWIN_ROCK: gg_gxwin = ref;
             }
          1: ! stream
          2: ! fileref
        }
    }
    if (phase == 2) { ! Update our objects.
         MakeWindow();
  #Ifndef NOSOUND;            
         id = glk_schannel_iterate(0, gg_arguments);
         while (id) {
            switch (gg_arguments-->0) {
               GG_MUSICCHAN_ROCK: gg_musicchan = id;
           }
           id = glk_schannel_iterate(id, gg_arguments);
         }
         if (~~musicflag) current_music = 0;
         MyRestartMusicChannel();         
  #Endif;
    }
    id=id;
    return;
];

#IfNot; ! TARGET_ZCODE;

! --- Styles wrappers for zcode

[ SetStyle x;
    switch(x) {
      SROMAN: style roman;
      SBOLD: style bold;
      SITALIC: style underline;
      SSOFT: style roman;
      SREVERSE: style reverse;
    }
];

#Endif;

Stub CloseWindow 0;
Stub MakeWindow 0;

Stub SetImage 1;
Stub ResetImage 1;
Stub CheckImage 1;

Stub LeftPictureCheck 1;
Stub RightPictureCheck 1;
Stub AbovePictureCheck 1;
Stub LeftPicture 1;
Stub RightPicture 1;
Stub AbovePicture 1;
Stub NoPicture 0;
Stub NoPictureNewLoc 0;

Stub MusicSwitch 0;
Stub MusicOff 0;
Stub MusicOn 0;

Stub PlayMusic 1;
Stub FadeToMusic 1;
Stub StopMusic 0;
Stub FadeOutMusic 0;

![ __GlkWarningSquasher;
!  CloseWindow(); MakeWindow();
!  SetImage(); CheckImage(); ResetImage();
!  LeftPictureCheck(); RightPictureCheck(); AbovePictureCheck(); 
!  LeftPicture(); RightPicture(); AbovePicture(); NoPicture(); NoPictureNewLoc();
!  __GlkWarningSquasher();
!];
